//'+--------------------------------------------------------------------------+
//'|                                                                          |
//'|                                Aquarium                                  |
//'|                                                                          |
//'|          GDImage control displaying multiple transparent layers          |
//'|                                                                          |
//'+--------------------------------------------------------------------------+
//'|                                                                          |
//'|                         Author Patrice TERRIER                           |
//'|                         copyright(c) 2002-2024                           |
//'|                           www.zapsolution.com                            |
//'|                        pterrier@zapsolution.com                          |
//'|                                                                          |
//'|                            www.objreader.com.............................|
//'|                           objreader@gmail.com                            |
//'|                                                                          |
//'+--------------------------------------------------------------------------+
//'|                  Project started on : 06-27-2023 (MM-DD-YYYY)            |
//'|                        Last revised : 08-01-2024 (MM-DD-YYYY)            |
//'+--------------------------------------------------------------------------+

#include <windows.h>
#include <commctrl.h>
#include <time.h>

#include "Tools.h"
#include "WinLIFT.h"
#include "GDImage.h"

#pragma warning(disable: 4312)  // remove warning C4312:  'type cast': conversion from 'long' to 'HMENU' of greater size

#pragma comment(lib, "WinLIFT64.lib")
#pragma comment(lib, "GDImage64.lib")

// Enable tooltip
#pragma comment(linker,"\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

constexpr auto IDI_ICON1         = 101;

const int CLIENT_WIDTH           = 608;
const int CLIENT_HEIGHT          = 513;

const int CTRL_WIDTH             = 606;
const int CTRL_HEIGHT            = 452;

const int TIMER_DELAY            = 60;

const int IDC_CLOSE              = 101;
const int IDC_TASKBAR            = 104;
const int IDC_LIBBACK            = 105;
const int IDC_GDIMAGE            = 106;
const int IDC_TIMER              = 107;
const int IDC_LIST               = 108;

const int ID_BACK                = 1;
const int ID_BUBBLE              = 2;
const int ID_FISHL               = 3;
const int ID_FISHR               = 4;
const int ID_LEFT                = 5;
const int ID_RIGHT               = 6;
const int ID_IMAGE               = 7;
const int ID_ICON                = 8;

typedef struct {
    HWND  hMain;
    HWND  hGDImage;

    long  UseID;
    BYTE  UseFrame;
    long  ToRight;
    long  Flip;
    DWORD FishTick;
} PROP;

PROP gP;

static long DetectMultiframe(long w, long h) {
    long nCount = w / h;
    if ((nCount * h) != w) {
        nCount = 1;
    } else {
        w = h;
    }
    return nCount;
}

static void AddChildControls() {

    // Create the GDImage graphic control
    gP.hGDImage = CreateWindowEx(0, L"ZIMAGECTRL", NULL, WS_CHILD | WS_VISIBLE, 1, 1, CTRL_WIDTH, CTRL_HEIGHT, gP.hMain, (HMENU) IDC_GDIMAGE, zInstance(), NULL);
    if (IsWindow(gP.hGDImage)) {
        DWORD UseBackColor = ZD_ARGB(64, 0,100,128); // Use a transparent cyan color for the background.
        ZI_SetProperty(gP.hGDImage, ZI_GradientTop, UseBackColor);
        ZI_SetProperty(gP.hGDImage, ZI_GradientBottom, UseBackColor);
    }

    HWND hCtrl = 0;
    HBITMAP hBitmap = 0;
    long w = 0, h = 0;
    long nCount = 0;

    hBitmap = ZI_CreateBitmapFromFile(Resource(L"background.png"), w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 0, 0, hBitmap, ZD_ColorARGB(255, 0), ID_BACK, ZS_VISIBLE);
    ZD_SetObjectHidden(ID_BACK, ZS_INACTIVE); // Do not take focus, mouse messages are sent to the underlaying sprites.

    // Create a disabled transparent layer (ZS_INACTIVE), 
    // that would give the focus to the underlaying sprites.
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"bubbles.png"), w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, CTRL_WIDTH - w, 0, hBitmap, ZD_ColorARGB(255, 0), ID_BUBBLE, ZS_VISIBLE);
    ZD_SetObjectHidden(ID_BUBBLE, ZS_INACTIVE); // Do not take focus, mouse messages are sent to the underlaying sprites.

    // Here we use a multiframe bitmap rather than a PNG animation
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"BlueL.png"), w, h);
    nCount = DetectMultiframe(w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, (CTRL_WIDTH - h) / 2, (long) (CTRL_HEIGHT - (h * 1.5f)), hBitmap, ZD_ColorARGB(255, 0), ID_FISHL, ZS_VISIBLE);
    ZD_SetObjectFrameCount(ID_FISHL, (BYTE) nCount);       // Save frame count
    ZD_SetObjectFrameToUse(ID_FISHL, 1, ZD_DRAW_DEFERRED); // Start on frame 1
    gP.UseID = ID_FISHL;

    // Here we use multiframe bitmap rather than PNG animation
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"BlueR.png"), w, h);
    nCount = DetectMultiframe(w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 0, 0, hBitmap, ZD_ColorARGB(255, 0), ID_FISHR, ZS_HIDDEN);
    ZD_SetObjectFrameCount(ID_FISHR, (BYTE) nCount);       // Save frame count
    ZD_SetObjectFrameToUse(ID_FISHR, 1, ZD_DRAW_DEFERRED); // Start on frame 1

    // Here we use multiframe bitmap rather than PNG animation
    hBitmap  = ZI_CreateBitmapFromFile(Resource(L"BlueTL.png"), w, h);
    nCount = DetectMultiframe(w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 0, 0, hBitmap, ZD_ColorARGB(255, 0), ID_LEFT, ZS_HIDDEN);
    ZD_SetObjectFrameCount(ID_LEFT, (BYTE) nCount);       // Save frame count
    ZD_SetObjectFrameToUse(ID_LEFT, 1, ZD_DRAW_DEFERRED); // Start on frame 1

    // Here we use multiframe bitmap rather than PNG animation
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"BlueTR.png"), w, h);
    nCount = DetectMultiframe(w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 0, 0, hBitmap, ZD_ColorARGB(255, 0), ID_RIGHT, ZS_HIDDEN);
    ZD_SetObjectFrameCount(ID_RIGHT, (BYTE) nCount);       // Save frame count
    ZD_SetObjectFrameToUse(ID_RIGHT, 1, ZD_DRAW_DEFERRED); // Start on frame 1

    // Create a disabled transparent layer (ZS_INACTIVE), 
    // that would give the focus to the underlaying sprites.
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"aero_glass.png"), w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 0, 0, hBitmap, ZD_ColorARGB(128, 0), ID_IMAGE, ZS_VISIBLE);
    ZD_SetObjectHidden(ID_IMAGE, ZS_INACTIVE); // Do not take focus, mouse messages are sent to the underlaying sprites.

    // Add the tutorial icon, so far it does nothing...
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"icon.png"), w, h);
    ZD_DrawBitmapToCtrl(gP.hGDImage, 8, 8, hBitmap, ZD_ColorARGB(255, 0), ID_ICON, ZS_VISIBLE);
    ZD_SetObjectHidden(ID_ICON, ZS_INACTIVE); // Do not take focus, mouse messages are sent to the underlaying sprites.

    // We use a WinLIFT static image to create the window frame,
    // WARNING this is not a real skinned form!!!
    // Use the bottom taskbar to move the window around.
    hCtrl = skStaticImage(gP.hMain, Resource(L"frame.png"), -2, 0, 612, 516, IDC_TASKBAR);

    hCtrl = skButtonImage(gP.hMain, Resource(L"BTN_Close.png"), 278, 458, IDC_CLOSE, 0);
    skCreateToolTip(hCtrl, L"Close");
}

static void PlayBubble() {
    PlaySound(Resource(L"bubble.wav"), NULL, SND_FILENAME | SND_ASYNC);
}

static void Animate() {

    POINT lp = { 0 };
    RECT r = { 0 };
    long x = 0, y = 0, w = 0, h = 0, nStepX = 0, nStepY = 0, nID = 0;

    DWORD nTickCount = (DWORD) GetTickCount64();
    if (gP.FishTick == 0) gP.FishTick = nTickCount + TIMER_DELAY;
    if (nTickCount > gP.FishTick) {

        nID = ID_BACK;
        ZD_GetObjectXY(nID, x, y);
        ZD_GetObjectBound(ID_BACK, w, h);
        if (gP.Flip == 0) {
            if (x > CTRL_WIDTH - w) {
                x -= 1;
            } else {
                x += 1;
                gP.Flip = -1;
            }
        } else {
            if (x < 0) {
                x += 1;
            } else {
                x = 0;
                gP.Flip = 0;
            }
        }
        ZD_SetObjectXY(nID, x, y, FALSE);

        gP.FishTick = 0;
        GetClientRect(gP.hGDImage, &r);
        Randomize();

        // Bubble
        ZD_GetObjectXY(ID_BUBBLE, x, y);
        ZD_GetObjectBound(ID_BUBBLE, w, h);
        y -= 4;
        if (y < -h) { 
            y = r.bottom;
            PlayBubble();
        }
        ZD_SetObjectXY(ID_BUBBLE, x, y, FALSE);

        // Fish
        ZD_GetObjectXY(gP.UseID, x, y);
        ZD_GetObjectBound(gP.UseID, w, h); w = h;
        nStepX = 4;
        nStepY = (long)((Random(0,1) - 0.5f) * 1.5f);
        lp.y = y + nStepY;
        if (gP.ToRight) {
            if ((x + w) < r.right) {
                lp.x = x + nStepX;
            } else {
                ZD_SetObjectVisibility(gP.UseID, ZS_HIDDEN);
                gP.UseID = ID_RIGHT;
                gP.UseFrame = 0;
                gP.ToRight = 0;
                lp.x = x - nStepX;
                ZD_SetObjectVisibility(gP.UseID, ZS_VISIBLE);
            }
        } else {
            if (x > 0) {
                lp.x = x - nStepX;
            } else {
                ZD_SetObjectVisibility(gP.UseID, ZS_HIDDEN);
                gP.UseID = ID_LEFT;
                gP.UseFrame = 0;
                gP.ToRight = -1;
                lp.x = x + nStepX;
                ZD_SetObjectVisibility(gP.UseID, ZS_VISIBLE);
            }
        }

        gP.UseFrame += 1;
        if (gP.UseFrame > ZD_GetObjectFrameCount(gP.UseID)) {

            if (gP.UseID == ID_LEFT) {
                ZD_SetObjectVisibility(gP.UseID, ZS_HIDDEN);
                gP.UseID = ID_FISHR;
                ZD_SetObjectVisibility(gP.UseID, ZS_VISIBLE);
            } else if (gP.UseID == ID_RIGHT) {
                ZD_SetObjectVisibility(gP.UseID, ZS_HIDDEN);
                gP.UseID = ID_FISHL;
                ZD_SetObjectVisibility(gP.UseID, ZS_VISIBLE);
            }

            gP.UseFrame = 1;
        }
        if (gP.UseID) {
            ZD_SetObjectXY(gP.UseID, lp.x, lp.y, FALSE);
            ZD_SetObjectFrameToUse(gP.UseID, gP.UseFrame, ZS_VISIBLE);
        }
    }
}

static LRESULT CALLBACK WndProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    long wmId = 0, wmEvent = 0;

    switch (uMsg) {
    case WM_CREATE:
        SetTimer(hWnd, IDC_TIMER, 0, 0); // 60 FPS
        break;

    case WM_COMMAND:
        wmId    = LOINT(wParam);
        wmEvent = HIINT(wParam);
        if (wmId == IDC_CLOSE) PostMessage(hWnd, WM_SYSCOMMAND, SC_CLOSE, 0);
        break;

    case WM_TIMER:
        Animate();
        break;

    case WM_NCHITTEST:
        return HTCAPTION;

    case WM_DESTROY:
        KillTimer(hWnd, IDC_TIMER);
        PostQuitMessage(0);
        return 0;
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

int WINAPI wWinMain(_In_ HINSTANCE hInstance, _In_opt_ HINSTANCE hPrevInstance, _In_ LPWSTR lpCmdLine, _In_ int nCmdShow) {

    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);

    // Declaration of local variables
    long nRet = 0, IsInitialized = 0;
    WNDCLASSEX wcx = { 0 };
    WCHAR szClass[] = L"Aqua64";

    // Detect process without using a Mutex
    HWND hFound = FindWindow(szClass, NULL);
    if (hFound) {
        if (IsIconic(hFound)) { ShowWindow(hFound, SW_RESTORE); }
        SetForegroundWindow(hFound);
        return nRet;
    }

    // Register the main window class
    wcx.cbSize = sizeof(WNDCLASSEX);
    IsInitialized = GetClassInfoEx(hInstance, szClass, &wcx);
    if (!IsInitialized) {
        wcx.style         = CS_HREDRAW | CS_VREDRAW;
        wcx.lpfnWndProc   = WndProc;
        wcx.cbClsExtra    = 0;
        wcx.cbWndExtra    = 0;
        wcx.hInstance     = hInstance;
        wcx.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
        wcx.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = NULL;
        wcx.lpszMenuName  = NULL;
        wcx.lpszClassName = szClass;
        wcx.hIconSm       = wcx.hIcon;
        if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }

    if (IsInitialized) {

        RECT r = { 0 };
        DWORD dwExStyle = 0;
        DWORD dwStyle = WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
        // Adjust the window to use the client size.
        SetRect(&r, 0, 0, CLIENT_WIDTH, CLIENT_HEIGHT);
        AdjustWindowRectEx(&r, dwStyle, FALSE, dwExStyle);
        long x = max((GetSystemMetrics(SM_CXSCREEN) - rWidth(r)) / 2, 0);
        long y = max((GetSystemMetrics(SM_CYSCREEN) - rHeight(r)) / 2, 0);

        gP.hMain = CreateWindowEx(dwExStyle, szClass, szClass, dwStyle, x, y,  rWidth(r), rHeight(r), NULL, NULL, hInstance, NULL);
        if (gP.hMain) {

            // WARNING, we use a very specific Skin for this application
            // it couldn't be used to skin a regular window.
            if (skInitEngine(Resource(L"skin.sks"), NULL)) {
                skSkinWindow(gP.hMain, NULL);

                AddChildControls();

                ShowWindow(gP.hMain, nCmdShow);
                SetForegroundWindow(gP.hMain);
                SetFocus(gP.hMain);

                // Play bubbles wav
                PlayBubble();

                MSG msg = { 0 };
                while (GetMessage(&msg, NULL, 0, 0)) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                nRet = (long) msg.wParam;
            }
        }
    }
    return nRet;
}
